"""
The Jacobian Inverse Transposed for axiparallel cube grids is diagonal.
We enforce this through a post-processing transformation (instead of hacking it into UFL)
"""

from dune.perftool.ufl.transformations import ufl_transformation
from ufl.algorithms import MultiFunction
from ufl.classes import (Indexed,
                         JacobianInverse,
                         MultiIndex,
                         Product,
                         Restricted,
                         )


class LocalDiagonalJITReplacer(MultiFunction):
    """Make JacobianInverse diagonal and simplify tree

    Search for:

    (IndexSum)-L->(Product)-L->(Indexed)-L->(JacobianInverse)

    and replace it with

    (Product)-L->(Indexed)-L->(JacobianInverse)

    where the index of the initial IndexSum is replaced by the second
    index of the JacobianInverse.

    Local means: Instead of replacing all appearances of this Index in
    the graph only this subtree is changed.
    """
    call = MultiFunction.__call__

    def __call__(self, o):
        self.local_replacemap = {}
        return self.call(o)

    def expr(self, o):
        return self.reuse_if_untouched(o, *tuple(self.call(op) for op in o.ufl_operands))

    def multi_index(self, o):
        return MultiIndex(tuple(self.local_replacemap.get(i, i) for i in o))

    def indexed(self, o):
        if self.local_replacemap:
            if isinstance(o.ufl_operands[1], MultiIndex):
                return self.reuse_if_untouched(o, *tuple(self.call(op) for op in o.ufl_operands))
        else:
            return self.reuse_if_untouched(o, *tuple(self.call(op) for op in o.ufl_operands))


    def index_sum(self, o):
        # Check if we want to replace an index here and store it in
        # local_replacemap
        e, i = o.ufl_operands
        if isinstance(e, Product):
            p1, p2 = e.ufl_operands
            if isinstance(p1, Indexed):
                p, ii = p1.ufl_operands
                if isinstance(p, Restricted):
                    p = p.ufl_operands[0]
                if isinstance(p, JacobianInverse):
                    assert(i[0] == ii[0])
                    assert(self.local_replacemap == {})
                    self.lock_replacemap = True
                    self.local_replacemap[ii[0]] = ii[1]

        if self.local_replacemap == {}:
            return self.reuse_if_untouched(o, *tuple(self.call(op) for op in o.ufl_operands))
        else:
            # Go through subtrees and replace indices if necessary
            e = self.call(e)

            # Clear replacemap
            self.local_replacemap.clear()
            return e


@ufl_transformation(name='axiparallel')
def diagonal_jacobian(e):
    """
    This transformations can generically be described as:
    \sum_k J^{(+/-)}_{k,i}x_k ==> J^{(+/-)}_{i,i}x_i
    """
    return LocalDiagonalJITReplacer()(e)
